#!/usr/bin/env python
# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/cp2k/aiida-cp2k      #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

import os
import sys
import time
import subprocess
import numpy as np
from os import path

import ase.build

from aiida import load_dbenv, is_dbenv_loaded
from aiida.backends import settings
if not is_dbenv_loaded():
    load_dbenv(profile=settings.AIIDADB_PROFILE)

from aiida.common.example_helpers import test_and_get_code
from aiida.orm.data.structure import StructureData
from aiida.orm.data.parameter import ParameterData
from aiida.orm.data.singlefile import SinglefileData

#===============================================================================
def main():
    if len(sys.argv) != 2:
        print("Usage: test.py <code_name>")
        sys.exit(1)

    codename = sys.argv[1]
    code = test_and_get_code(codename, expected_code_type='cp2k')
    test_no_structure_data(code)
    test_energy_mm(code)
    test_energy_dft(code)
    test_geo_opt_dft(code)
    print("All tests passed :-)")
    sys.exit(0)

#===============================================================================
def test_energy_mm(code):
    print("Testing CP2K ENERGY on H2O (MM) ...")

    # calc object
    calc = code.new_calc()

    # parameters
    # based on cp2k/tests/Fist/regtest-1-1/water_1.inp
    parameters = ParameterData(dict={
            'FORCE_EVAL':{
                'METHOD': 'fist',
                'MM': {
                    'FORCEFIELD': {
                        'PARM_FILE_NAME': 'water.pot',
                        'PARMTYPE': 'CHM',
                        'CHARGE':[
                            {'ATOM':'O', 'CHARGE': -0.8476},
                            {'ATOM':'H', 'CHARGE': 0.4238},]
                    },
                    'POISSON': {'EWALD':{
                        'EWALD_TYPE':'spme',
                        'ALPHA': 0.44,
                        'GMAX': 24,
                        'O_SPLINE': 6
                    }}
                }
            },
            'GLOBAL': {
                'CALLGRAPH': 'master',
                'CALLGRAPH_FILE_NAME': 'runtime'
            }
    })
    calc.use_parameters(parameters)

    # force field
    with open("water.pot", "w") as f:
        f.write("""BONDS
H    H       0.000     1.5139
O    H     450.000     0.9572

ANGLES
H    O    H      55.000   104.5200

DIHEDRALS

IMPROPER

NONBONDED
H      0.000000  -0.046000     0.224500
O      0.000000  -0.152100     1.768200

HBOND CUTHB 0.5

END""")
    water_pot = SinglefileData(file=path.abspath("water.pot"))
    calc.use_file(water_pot, linkname="water_pot")

    # structure
    atoms = ase.build.molecule('H2O')
    atoms.center(vacuum=10.0)
    structure = StructureData(ase=atoms)
    calc.use_structure(structure)

    # resources
    calc.set_max_wallclock_seconds(3*60) # 3 min
    calc.set_resources({"num_machines": 1})

    settings = ParameterData(dict={'additional_retrieve_list':["runtime.callgraph"]})
    calc.use_settings(settings)
    #calc.submit_test()
    #return

    # store and submit
    calc.store_all()
    calc.submit()
    print "submitted calculation: UUID=%s, pk=%s"%(calc.uuid,calc.dbnode.pk)

    wait_for_calc(calc)

    # check energy results
    expected_energy = 0.146927412614e-3
    if abs(calc.res.energy - expected_energy) < 1e-10:
        print "OK, energy has the expected value"
    else:
        print "ERROR!"
        print "Expected energy value: {}".format(expected_energy)
        print "Actual energy value: {}".format(calc.res.energy)
        sys.exit(3)

    # check if callgraph is there
    if "runtime.callgraph" in calc.get_retrieved_node().get_folder_list():
        print "OK, callgraph file was retrived"
    else:
        print "ERROR!"
        print "Callgraph file was not retrieved."
        sys.exit(3)

#===============================================================================
def test_energy_dft(code):
    print("Testing CP2K ENERGY on H2O (DFT)...")

    # calc object
    calc = code.new_calc()

    # structure
    atoms = ase.build.molecule('H2O')
    atoms.center(vacuum=2.0)
    structure = StructureData(ase=atoms)
    calc.use_structure(structure)

    # parameters
    parameters = ParameterData(dict={'FORCE_EVAL':get_force_eval()})
    calc.use_parameters(parameters)

    # resources
    calc.set_max_wallclock_seconds(3*60) # 3 min
    calc.set_resources({"num_machines": 1})

    # store and submit
    calc.store_all()
    calc.submit()
    print "submitted calculation: UUID=%s, pk=%s"%(calc.uuid,calc.dbnode.pk)

    wait_for_calc(calc)

    # check results
    expected_energy = -17.1566368539
    if abs(calc.res.energy - expected_energy) < 1e-10:
        print "OK, energy has the expected value"
    else:
        print "ERROR!"
        print "Expected energy value: {}".format(expected_energy)
        print "Actual energy value: {}".format(calc.res.energy)
        sys.exit(3)

#===============================================================================
def test_geo_opt_dft(code):
    print("Testing CP2K GEO_OPT on H2 (DFT)...")

    # calc object
    calc = code.new_calc()

    # structure
    atoms = ase.build.molecule('H2')
    atoms.center(vacuum=2.0)
    vec = atoms.positions[0,:] - atoms.positions[1,:]
    structure = StructureData(ase=atoms)
    calc.use_structure(structure)

    # parameters
    parameters = ParameterData(dict={
        'GLOBAL': {
            'RUN_TYPE': 'GEO_OPT',
        },
        'FORCE_EVAL': get_force_eval()
    })
    calc.use_parameters(parameters)

    # resources
    calc.set_max_wallclock_seconds(3*60) # 3 min
    calc.set_resources({"num_machines": 1})

    # store and submit
    calc.store_all()
    calc.submit()
    print "submitted calculation: UUID=%s, pk=%s"%(calc.uuid,calc.dbnode.pk)

    wait_for_calc(calc)

    # check results
    expected_energy = -1.14009973333
    if abs(calc.res.energy - expected_energy) < 1e-10:
        print "OK, energy has the expected value"
    else:
        print "ERROR!"
        print "Expected energy value: {}".format(expected_energy)
        print "Actual energy value: {}".format(calc.res.energy)
        sys.exit(3)

    expected_dist = 0.736125211
    dist = calc.out.output_structure.get_ase().get_distance(0, 1)
    if abs(dist - expected_dist) < 1e-7:
        print "OK, H-H distance has the expected value"
    else:
        print "ERROR!"
        print "Expected dist value: {}".format(expected_dist)
        print "Actual dist value: {}".format(dist)
        sys.exit(3)

#===============================================================================
def test_no_structure_data(code):
    print("Testing CP2K ENERGY on H2 (DFT) without StructureData...")

    # calc object
    calc = code.new_calc()

    # structure directly included in parameters
    force_eval = get_force_eval()
    force_eval['SUBSYS']['CELL'] = {'ABC': '4.0   4.0   4.75'}
    force_eval['SUBSYS']['COORD'] = {' ': ['H    2.0   2.0   2.737166',
                                           'H    2.0   2.0   2.000000']}

    # parameters
    parameters = ParameterData(dict={'FORCE_EVAL':force_eval})
    calc.use_parameters(parameters)

    # resources
    calc.set_max_wallclock_seconds(3*60) # 3 min
    calc.set_resources({"num_machines": 1})

    # store and submit
    calc.store_all()
    calc.submit()
    print "submitted calculation: UUID=%s, pk=%s"%(calc.uuid,calc.dbnode.pk)

    wait_for_calc(calc)

    # check results
    expected_energy = -1.14005678487
    if abs(calc.res.energy - expected_energy) < 1e-10:
        print "OK, energy has the expected value"
    else:
        print "ERROR!"
        print "Expected energy value: {}".format(expected_energy)
        print "Actual energy value: {}".format(calc.res.energy)
        sys.exit(3)

#===============================================================================
def get_force_eval():
    return {
            'METHOD': 'Quickstep',
            'DFT': {
                'BASIS_SET_FILE_NAME': 'BASIS_MOLOPT',
                'QS': {
                    'EPS_DEFAULT': 1.0e-12,
                    'WF_INTERPOLATION': 'ps',
                    'EXTRAPOLATION_ORDER': 3,
                },
                'MGRID': {
                    'NGRIDS': 4,
                    'CUTOFF': 280,
                    'REL_CUTOFF': 30,
                },
                'XC': {
                    'XC_FUNCTIONAL': {
                        '_': 'LDA',
                    },
                },
                'POISSON': {
                    'PERIODIC': 'none',
                    'PSOLVER': 'MT',
                },
            },
            'SUBSYS': {
                'KIND': [
                    {'_': 'O', 'BASIS_SET': 'DZVP-MOLOPT-SR-GTH', 'POTENTIAL': 'GTH-LDA-q6' },
                    {'_': 'H', 'BASIS_SET': 'DZVP-MOLOPT-SR-GTH', 'POTENTIAL': 'GTH-LDA-q1' },
                ],
            },
        }

#===============================================================================
def wait_for_calc(calc, timeout_secs=5*60.0):
    print "Wating for end of execution..."
    start_time = time.time()
    exited_with_timeout = True
    while time.time() - start_time < timeout_secs:
        time.sleep(15) # Wait a few seconds
        # print some debug info, both for debugging reasons and to avoid
        # that the test machine is shut down because there is no output
        print "#"*78
        print "####### TIME ELAPSED: {} s".format(time.time() - start_time)
        print "#"*78
        print "Output of 'verdi calculation list':"
        try:
            cmd = ["verdi", "calculation", "list"]
            print subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            print "Note: the command failed, message: {}".format(e.message)
        if calc.has_finished():
            print "Calculation terminated its execution"
            exited_with_timeout = False
            break

    # check results
    if exited_with_timeout:
        print "Timeout!! Calculation did not complete after {} seconds".format(
            timeout_secs)
        sys.exit(2)

#===============================================================================
if __name__ == "__main__":
    main()

#EOF



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/cp2k/aiida-cp2k      #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from aiida.orm.calculation.job import JobCalculation
from aiida.common.utils import classproperty
from aiida.orm.data.structure import StructureData
from aiida.orm.data.parameter import ParameterData
from aiida.orm.data.singlefile import SinglefileData
from aiida.orm.data.remote import RemoteData
from aiida.common.datastructures import CalcInfo, CodeInfo
from aiida.common.exceptions import InputValidationError
import os

class Cp2kCalculation(JobCalculation):
    """
    This is a Cp2kCalculation, subclass of JobCalculation, to prepare input for an
    ab-inition Cp2kCalculation.
    For information on CP2K, refer to: https://www.cp2k.org
    """

    #---------------------------------------------------------------------------
    def _init_internal_params(self):
        """
        Set parameters of instance
        """
        super(Cp2kCalculation, self)._init_internal_params()
        self._INPUT_FILE_NAME  = 'aiida.inp'
        self._OUTPUT_FILE_NAME = 'aiida.out'
        self._DEFAULT_INPUT_FILE  = self._INPUT_FILE_NAME
        self._DEFAULT_OUTPUT_FILE = self._OUTPUT_FILE_NAME
        self._COORDS_FILE_NAME = 'aiida.coords.pdb'
        self._PROJECT_NAME = 'AIIDA'
        self._TRAJ_FILE_NAME = self._PROJECT_NAME+'-pos-1.xyz'
        self._default_parser = 'cp2k'

    #---------------------------------------------------------------------------
    @classproperty
    def _use_methods(cls):
        """
        Extend the parent _use_methods with further keys.
        This will be manually added to the _use_methods in each subclass
        """
        retdict = JobCalculation._use_methods
        retdict.update({
            "structure": {
               'valid_types': StructureData,
               'additional_parameter': None,
               'linkname': 'structure',
               'docstring': "Choose the input structure to use",
               },
            "settings": {
               'valid_types': ParameterData,
               'additional_parameter': None,
               'linkname': 'settings',
               'docstring': "Use an additional node for special settings",
               },
            "parameters": {
               'valid_types': ParameterData,
               'additional_parameter': None,
               'linkname': 'parameters',
               'docstring': "Use a node that specifies the input parameters for the namelists",
               },
            "parent_folder": {
               'valid_types': RemoteData,
               'additional_parameter': None,
               'linkname': 'parent_calc_folder',
               'docstring': "Use a remote folder as parent folder (for restarts and similar)",
               },
            "file": {
               'valid_types': SinglefileData,
               'additional_parameter': "linkname",
               'linkname': cls._get_linkname_file,
               'docstring': "Use files to provide additional parameters",
               },
            })
        return retdict

    #---------------------------------------------------------------------------
    @classmethod
    def _get_linkname_file(cls, linkname):
        return(linkname)

    #---------------------------------------------------------------------------
    def _prepare_for_submission(self, tempfolder, inputdict):
        """
        This is the routine to be called when you want to create
        the input files and related stuff with a plugin.

        :param tempfolder: a aiida.common.folders.Folder subclass where
                           the plugin should put all its files.
        :param inputdict: a dictionary with the input nodes, as they would
                be returned by get_inputdata_dict (without the Code!)
        """

        params, structure, code, settings, local_copy_list = self._verify_inlinks(inputdict)

        # write cp2k input file
        inp = Cp2kInput(params)
        if structure is not None:
            struct_fn = tempfolder.get_abs_path(self._COORDS_FILE_NAME)
            structure.get_ase().write(struct_fn, format="proteindatabank")
            for i, a in enumerate('ABC'):
                val = '{:<15} {:<15} {:<15}'.format(*structure.cell[i])
                inp.add_keyword('FORCE_EVAL/SUBSYS/CELL/'+a, val)
            inp.add_keyword("FORCE_EVAL/SUBSYS/TOPOLOGY/COORD_FILE_NAME", self._COORDS_FILE_NAME)
            inp.add_keyword("FORCE_EVAL/SUBSYS/TOPOLOGY/COORD_FILE_FORMAT", "pdb")
        inp.add_keyword("GLOBAL/PROJECT", self._PROJECT_NAME)
        inp_fn = tempfolder.get_abs_path(self._INPUT_FILE_NAME)
        with open(inp_fn, "w") as f:
            f.write(inp.render())

        # create code info
        codeinfo = CodeInfo()
        cmdline = settings.pop('cmdline', [])
        cmdline += ["-i", self._INPUT_FILE_NAME]
        codeinfo.cmdline_params = cmdline
        codeinfo.stdout_name = self._OUTPUT_FILE_NAME
        codeinfo.join_files = True
        codeinfo.code_uuid = code.uuid

        # create calc info
        calcinfo = CalcInfo()
        calcinfo.stdin_name = self._INPUT_FILE_NAME
        calcinfo.uuid = self.uuid
        calcinfo.cmdline_params = codeinfo.cmdline_params
        calcinfo.stdin_name = self._INPUT_FILE_NAME
        calcinfo.stdout_name = self._OUTPUT_FILE_NAME
        calcinfo.codes_info = [codeinfo]

        # file lists
        calcinfo.remote_symlink_list = []
        calcinfo.local_copy_list = local_copy_list
        calcinfo.remote_copy_list = []
        calcinfo.retrieve_list = [self._OUTPUT_FILE_NAME, self._TRAJ_FILE_NAME]
        calcinfo.retrieve_list += settings.pop('additional_retrieve_list', [])

        # check for left over settings
        if settings:
            msg = "The following keys have been found in the settings input node, "
            msg = "but were not understood: " + ",",join(settings_dict.keys())
            raise InputValidationError(msg)

        return calcinfo

    #---------------------------------------------------------------------------
    def _verify_inlinks(self, inputdict):
        # parameters
        try:
            params_node = inputdict.pop(self.get_linkname('parameters'))
        except KeyError:
            raise InputValidationError("No parameters specified for this calculation")
        if not isinstance(params_node, ParameterData):
            raise InputValidationError("parameters is not of type ParameterData")
        params = params_node.get_dict()

        # structure
        structure = inputdict.pop(self.get_linkname('structure'), None)
        if structure is not None:
            if not isinstance(structure, StructureData):
                raise InputValidationError("structure is not of type StructureData")

        # code
        try:
            code = inputdict.pop(self.get_linkname('code'))
        except KeyError:
            raise InputValidationError("No code specified for this calculation")

        # settings
        # ... if not provided fall back to empty dict
        settings_node = inputdict.pop(self.get_linkname('settings'), ParameterData())
        if not isinstance(settings_node, ParameterData):
            raise InputValidationError("settings, if specified, must be of type ParameterData")
        settings = settings_node.get_dict()

        # parent calc folder (not yet used)
        parent_calc_folder = inputdict.pop(self.get_linkname('parent_folder'), None)
        if parent_calc_folder is not None:
            if not isinstance(parent_calc_folder, RemoteData):
                raise InputValidationError("parent_calc_folder, if specified, must be of type RemoteData")

        # handle additional parameter files
        local_copy_list = []
        for k, v in inputdict.items():
            if isinstance(v, SinglefileData):
                inputdict.pop(k)
                local_copy_list.append( (v.get_file_abs_path(), v.filename) )

        if inputdict:
            msg = "The following input data nodes are unrecognized: {}".format(inputdict.keys())
            raise InputValidationError(msg)

        return(params, structure, code, settings, local_copy_list)


#===============================================================================
class Cp2kInput(object):
    def __init__(self, params):
        self.params = params

    #---------------------------------------------------------------------------
    def add_keyword(self, kwpath, value, params=None):
        self._add_keyword_low(kwpath.split("/"), value, self.params)

    #---------------------------------------------------------------------------
    def _add_keyword_low(self, kwpath, value, params):
        if len(kwpath)==1:
            params[kwpath[0]] = value
        elif kwpath[0] not in params.keys():
            new_subsection = {}
            params[kwpath[0]] = new_subsection
            self._add_keyword_low(kwpath[1:], value, new_subsection)
        else:
            self._add_keyword_low(kwpath[1:], value, params[kwpath[0]])

    #---------------------------------------------------------------------------
    def render(self):
        output = ["!!! Generated by AiiDA !!!"]
        self._render_section(output, self.params)
        return "\n".join(output)

    #---------------------------------------------------------------------------
    def _render_section(self, output, params, indent=0):
        """
        It takes a dictionary and recurses through.

        For key-value pair it checks whether the value is a dictionary and prepends the key with &
        It passes the valued to the same function, increasing the indentation
        If the value is a list, I assume that this is something the user wants to store repetitively
        eg:
            dict['KEY'] = ['val1', 'val2']
            ===>
            KEY val1
            KEY val2

            or

            dict['KIND'] = [{'_': 'Ba', 'ELEMENT':'Ba'},
                            {'_': 'Ti', 'ELEMENT':'Ti'},
                            {'_': 'O', 'ELEMENT':'O'}]
            ====>
                  &KIND Ba
                     ELEMENT  Ba
                  &END KIND
                  &KIND Ti
                     ELEMENT  Ti
                  &END KIND
                  &KIND O
                     ELEMENT  O
                  &END KIND
        """

        for key, val in sorted(params.items()):
            if key.upper() != key:
                raise InputValidationError("CP2K keyword '%s' is not upper cased"%key)
            if key.startswith('@') or key.startswith('$'):
                raise InputValidationError("CP2K internal input preprocessor not supported in AiiDA")
            if isinstance(val, dict):
                output.append('{}&{} {}'.format(' '*indent, key, val.pop('_', '')))
                self._render_section(output, val, indent + 3)
                output.append('{}&END {}'.format(' '*indent, key))
            elif isinstance(val, list):
                for listitem in val:
                    self._render_section(output,  {key:listitem}, indent)
            elif isinstance(val, bool):
                val_str = '.true.'  if val else '.false.'
                output.append('{}{}  {}'.format(' '*indent, key, val_str))
            else:
                output.append('{}{}  {}'.format(' '*indent, key, val))


#EOF


# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/cp2k/aiida-cp2k      #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

import ase.io

from aiida.parsers.parser import Parser
from aiida.orm.data.parameter import ParameterData
from aiida.orm.data.structure import StructureData
from aiida.parsers.exceptions import OutputParsingError
from aiida_cp2k.calculations import Cp2kCalculation

class Cp2kParser(Parser):
    """
    Parser for the output of CP2K.
    """

    #---------------------------------------------------------------------------
    def __init__(self, calc):
        """
        Initialize the instance of Cp2kParser
        """
        super(Cp2kParser, self).__init__(calc)

        # check for valid input
        if not isinstance(calc, Cp2kCalculation):
            raise OutputParsingError("Input calc must be a Cp2kCalculation")

    #---------------------------------------------------------------------------  
    def parse_with_retrieved(self, retrieved):
        """
        Receives in input a dictionary of retrieved nodes.
        Does all the logic here.
        """
        out_folder = retrieved[self._calc._get_linkname_retrieved()]

        new_nodes_list = []
        self._parse_stdout(out_folder, new_nodes_list)
        self._parse_trajectory(out_folder, new_nodes_list)

        return True, new_nodes_list

    #---------------------------------------------------------------------------
    def _parse_stdout(self, out_folder, new_nodes_list):
        fn = self._calc._OUTPUT_FILE_NAME
        if fn not in out_folder.get_folder_list():
            raise OutputParsingError("Cp2k output file not retrieved")

        result_dict = {}
        abs_fn = out_folder.get_abs_path(fn)
        with open(abs_fn, "r") as f:
            for line in f.readlines():
                if line.startswith(' ENERGY| '):
                    result_dict['energy'] = float(line.split()[8])
                    result_dict['energy_units'] = "a.u."

        if 'energy' not in result_dict.keys():
            raise OutputParsingError("Cp2k output file without final energy")

        pair = (self.get_linkname_outparams(), ParameterData(dict=result_dict))
        new_nodes_list.append(pair)

    #---------------------------------------------------------------------------
    def _parse_trajectory(self, out_folder, new_nodes_list):
        fn = self._calc._TRAJ_FILE_NAME
        if fn not in out_folder.get_folder_list():
            return  # not every run type produces a trajectory

        abs_fn = out_folder.get_abs_path(fn)
        ase_struct = ase.io.read(abs_fn, index='-1')
        #TODO: parsing cell trajectories not yet implemented
        ase_struct.cell = self._calc.inp.structure.get_ase().cell
        pair = (self.get_linkname_outstructure(), StructureData(ase=ase_struct))
        new_nodes_list.append(pair)

    #---------------------------------------------------------------------------
    def get_linkname_outstructure(self):
        """
        Returns the name of the link to the output_structure
        """
        return 'output_structure'

    #---------------------------------------------------------------------------
    def get_linkname_outtrajectory(self):
        """
        Returns the name of the link to the output_trajectory.
        """
        return 'output_trajectory'

#EOF